import { groupBy, map, pipe, sum } from "ramda";
import { StrictOmit } from "ts-essentials";
import {
  UploadStatisticsDto,
  UploadStatisticsMimeRecord,
  UploadStatisticsMimeSegment,
  StorageStatisticsSegmentDto
} from "@upload-io/upload-shared-iso";

interface FileInfo {
  mime: string;
  size: number;
}

type Predicate<T> = (item: T) => boolean;
type MimeFilters = StrictOmit<UploadStatisticsMimeRecord<Predicate<FileInfo>>, "other">;
type EventsByMime = UploadStatisticsMimeRecord<FileInfo[]>;
type StatisticsByMime = UploadStatisticsMimeRecord<StorageStatisticsSegmentDto>;
type EventsFunc<T> = (events: FileInfo[]) => T;

const mimeFilters: MimeFilters = {
  audio: x => x.mime.startsWith("audio/"),
  image: x => x.mime.startsWith("image/"),
  video: x => x.mime.startsWith("video/"),
  document: x =>
    [
      "application/x-abiword",
      "application/msword",
      "application/vnd.ms-",
      "application/vnd.openxmlformats",
      "application/vnd.oasis.opendocument",
      "application/pdf"
    ].some(prefix => x.mime.startsWith(prefix))
};

const mimeFiltersEntries: Array<[UploadStatisticsMimeSegment, Predicate<FileInfo>]> = Object.entries(
  mimeFilters
).map(([k, v]) => [k as UploadStatisticsMimeSegment, v]);

export class UploadStatisticsCalculator {
  static calculateStatistics(newEvents: FileInfo[], valueMultiplier: 1 | -1): UploadStatisticsDto {
    const calculateSegment = (segment: FileInfo[]): StorageStatisticsSegmentDto => ({
      totalFileCount: valueMultiplier * segment.length,
      totalFileSize: valueMultiplier * sum(segment.map(x => x.size))
    });

    const groupByMime: EventsFunc<EventsByMime> = groupBy(event => {
      const match = mimeFiltersEntries.find(([_, filter]) => filter(event));
      if (match !== undefined) {
        return match[0];
      }
      return "other";
    });

    const calculateStatisticsForMime = map<EventsByMime, StatisticsByMime>((events: FileInfo[]) =>
      calculateSegment(events)
    );

    const newStatisticsMime: Partial<StatisticsByMime> = pipe(groupByMime, calculateStatisticsForMime)(newEvents);

    return {
      all: calculateSegment(newEvents),
      mime: {
        ...UploadStatisticsDto.empty.mime,
        ...newStatisticsMime
      }
    };
  }
}
